#!/usr/bin/env python
# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.
import cherrypy
import os
import socket
import logging
import mimetypes
import subprocess

import simplejson as json

import cmf.agent
import cmf.version
import cmf.log_search
import cmf.fencing_server

from cmf.agent_log_compressor import AgentLogCompressor
from cmf.template_util import render_html
from cmf.util import bytes_to_hex, validate_and_open
from cherrypy.lib.static import serve_fileobj

LOG = logging.getLogger(__name__)

class StatusServer(object):
  def __init__(self, agent, wakeup_fd):
    self.agent = agent
    self.host = cherrypy.config.get('server.socket_host')
    self.port = cherrypy.config.get('server.socket_port')
    self.hostname = socket.gethostname()
    self.wakeup_fd = wakeup_fd

  def __del__(self):
    if self.wakeup_fd is not None:
      os.close(self.wakeup_fd)
      self.wakeup_fd = None

  def _check_token(self):
    """
    Checks that the token given in the request cookie matches the agent's token.

    Throws appropriate error otherwise. Note that since the agent's traffic
    is not encrypted, anyone can sniff the token off the wire...
    """
    cookie = cherrypy.request.cookie
    if "token" not in cookie and self.agent.optional_token:
      return

    if "token" not in cookie or bytes_to_hex(self.agent.token) != cookie["token"].value:
      raise cherrypy.HTTPError(403, "Missing or invalid token.")

  def index(self):
    self._check_token()
    try:
      self.agent.lock.acquire()

      # Set the token in the cookie so that accessing internal pages would work
      # without having to proxy through the CM server
      cherrypy.response.cookie["token"] = bytes_to_hex(self.agent.token)
      return render_html("index.html",
          dict(version=cmf.version.VERSION, agent=self.agent))
    finally:
      self.agent.lock.release()
  index.exposed = True

  def _validate_and_open(self, path, no_links=True, restrict_to_parent_path=True, valid_owners=None):
    """
    Validates and opens a file, providing the agent directory as the parent
    directory if the caller wants to restric file access with the process
    directory. See util.validate_and_open.
    """
    parent_path = None
    if restrict_to_parent_path:
      parent_path = self.agent.dir

    try:
      return validate_and_open(path, parent_path, no_links, valid_owners)
    except Exception, e:
      raise cherrypy.HTTPError(500, str(e))

  def _serve_file(self, relativepath):
    """
    Serves a file relative to the agent process folder.
    """
    path = os.path.join(self.agent.dir, "process", relativepath)
    (content_type, _) = mimetypes.guess_type(path)

    f = self._validate_and_open(path)
    return serve_fileobj(f, content_type or "text/plain")

  def process(self, process_name, *args, **kw):
    """
    Handle requests for http://host/process/<process_name>/<op>
    where process_name is the process name and op is the particular operation
      op: view a file from the process_name private store (config directory)
    """
    self._check_token()
    op = args[0]
    type = args[1]
    args = args[1:]
    if op == "files" and type == "logs":
      return self._serve_file(os.path.join(process_name, *args))
    elif op == "files" and type == "host_statistics":
      return self._serve_file(os.path.join(process_name, args[1]))
    elif op == "files" and type == "inspector":
      return self._serve_file(os.path.join(process_name, "output.json"))
    elif op == "files" and type == "existing_configs":
      return self._serve_file(os.path.join(process_name, "configs.zip"))
    else:
      # unknown op
      msg = "Invalid request. Process: %s %s %s" % (process_name, op, args)
      raise cherrypy.HTTPError(404, msg)
  process.exposed = True

  def parse_search_logs_args(self,
                             start_time,
                             end_time,
                             log_paths,
                             log_types,
                             role_ids,
                             log_level,
                             search_timeout,
                             query,
                             role_result_limit,
                             total_byte_limit):

    # These are required args and should throw an exception if not specified.
    start_time = int(start_time)
    end_time = int(end_time)
    log_paths = log_paths.split(",")
    log_types = log_types.split(",")
    role_ids = role_ids.split(",")

    if log_level is not None:
      log_level = cmf.log_search.parse_log_level(log_level)

    if search_timeout is not None:
      search_timeout = int(search_timeout)

    if role_result_limit is not None:
      role_result_limit = int(role_result_limit)

    if total_byte_limit is not None:
      total_byte_limit = int(total_byte_limit)

    # No parsing needed for query

    return {
      "start_time": start_time,
      "end_time": end_time,
      "log_paths": log_paths,
      "log_types": log_types,
      "role_ids": role_ids,
      "log_level": log_level,
      "search_timeout": search_timeout,
      "query": query,
      "role_result_limit": role_result_limit,
      "total_byte_limit": total_byte_limit
    }

  def invoke_search_logs(self,
                         search_logs_func,
                         start_time=None,
                         end_time=None,
                         log_paths=None,
                         log_types=None,
                         role_ids=None,
                         log_level=None,
                         search_timeout=None,
                         query=None,
                         role_result_limit=None,
                         total_byte_limit=None):
    try:
      args = self.parse_search_logs_args(
        start_time, end_time, log_paths, log_types, role_ids, log_level,
        search_timeout, query, role_result_limit, total_byte_limit)

      result = search_logs_func(
        start_time=args["start_time"],
        end_time=args["end_time"],
        log_paths=args["log_paths"],
        log_types=args["log_types"],
        role_ids=args["role_ids"],
        log_level=args["log_level"],
        search_timeout=args["search_timeout"],
        query=args["query"],
        role_result_limit=args["role_result_limit"],
        total_byte_limit=args["total_byte_limit"],
        host=self.hostname,
        port=self.port)

      return result
    except Exception, e:
      return {
        "content_type": "application/json",
        "data": json.dumps(dict(
                  success=False,
                  messages=[e.args[0]], # the error message
                  host=self.host,
                  port=self.port))
      }

  def search_logs(self,
                  start_time=None,
                  end_time=None,
                  log_paths=None,
                  log_types=None,
                  role_ids=None,
                  log_level=None,
                  search_timeout=None,
                  query=None,
                  role_result_limit=None,
                  total_byte_limit=None):
    self._check_token()
    result = self.invoke_search_logs(
      search_logs_func=cmf.log_search.search_logs,
      start_time=start_time,
      end_time=end_time,
      log_paths=log_paths,
      log_types=log_types,
      role_ids=role_ids,
      log_level=log_level,
      search_timeout=search_timeout,
      query=query,
      role_result_limit=role_result_limit,
      total_byte_limit=total_byte_limit)

    cherrypy.response.headers["Content-Type"] = result["content_type"]
    return result["data"]

  search_logs.exposed = True

  def search_logs_compressed(self,
                             start_time=None,
                             end_time=None,
                             log_paths=None,
                             log_types=None,
                             role_ids=None,
                             log_level=None,
                             search_timeout=None,
                             query=None,
                             role_result_limit=None,
                             total_byte_limit=None):
    self._check_token()
    result = self.invoke_search_logs(
      search_logs_func=cmf.log_search.search_logs_compressed,
      start_time=start_time,
      end_time=end_time,
      log_paths=log_paths,
      log_types=log_types,
      role_ids=role_ids,
      log_level=log_level,
      search_timeout=search_timeout,
      query=query,
      role_result_limit=role_result_limit,
      total_byte_limit=total_byte_limit)

    f = self._validate_and_open(result["path"], restrict_to_parent_path=False)
    return serve_fileobj(f, result["content_type"])
  search_logs_compressed.exposed = True

  def log_with_context(self,
                       path=None,
                       offset=None,
                       direction=None,
                       num_events="100",
                       log_type=None):
    """
    Shows you a single log entry in the context of its sorrounding log entries
    Make a request like so:
      /log_in_context?path=/var/log/hadoop/datanode.log&offset=0&logType=LOG4J
    """
    self._check_token()
    try:
      if offset is not None:
        offset = int(offset)
      num_events = int(num_events)

      result = cmf.log_search.log_with_context(
          path,
          offset,
          direction,
          num_events,
          log_type,
          host=self.host,
          port=self.port)

      cherrypy.response.headers["Content-Type"] = result["content_type"]
      return result["data"]
    except Exception, e:
        cherrypy.response.headers["Content-Type"] = "application/json"
        return json.dumps(dict(
          success=False,
          message=str(e),
          usage=self.log_with_context.__doc__,
          host=self.host,
          port=self.port))
  log_with_context.exposed = True

  def is_impala_log(self, log_file):
    return (log_file.startswith("impalad") or
            log_file.startswith("impala-state-store"))

  def download_log(self, path=None, plaintext=False):
    """
    This function is used for downloading a log, either as an attachment
    or a text file. Which method is used is determined by the plaintext
    variable. By default, the file is downloaded as an attachment.
    """
    self._check_token()

    # As agent code runs as root, we don't want to follow all symlinks
    # as somone with lower privilege can create a symlink to a privileged
    # file. We will then end up with privilege escalation security hole
    # where agent running as root will allow download of this file.
    # To prevent this, we allow symlink resolution in directory part
    # of the path and file can be a symlink as long as it is in the same
    # directory. We can then safely open log files with O_NOFOLLOW to
    # prevent exploits.
    (log_dir, log_file) = os.path.split(path)

    new_log_file_path = None
    if self.is_impala_log(log_file):
      new_log_file_path = os.path.realpath(path)
      (log_dir, log_file) = os.path.split(new_log_file_path)
      if not self.is_impala_log(log_file):
        raise cherrypy.HTTPError (403, "path %s is not accessible." % (path,))
    else:
      log_dir = os.path.realpath(log_dir)
      new_log_file_path = os.path.join(log_dir, log_file)

    f = self._validate_and_open(new_log_file_path, restrict_to_parent_path=False)

    if plaintext:
      return serve_fileobj(f, "text/plain")
    else:
      return serve_fileobj(f, "application/x-download", "attachment")
  download_log.exposed = True

  def pstree(self):
    """
    Returns a pstree from the current process.
    """
    self._check_token()
    cherrypy.response.headers["Content-Type"] = "text/plain"
    p = subprocess.Popen(
      ["/usr/bin/pstree", "--show-pids", "--uid-changes",
       "--long", "--arguments", str(os.getpid()) ],
      stdin=None, stdout=subprocess.PIPE, close_fds=True, shell=False)
    return p.communicate()[0]
  pstree.exposed = True

  def logs(self):
    """
    Returns the last agent.BUFFER_SIZE characters of the agent's log. Newest
    entries are at the bottom.
    """
    self._check_token()
    cherrypy.response.headers["Content-Type"] = "text/plain"
    buf = cmf.agent.get_log_buffer()
    if buf:
      return str("\n".join(buf))
    else:
      return "No log buffer found."
  logs.exposed = True

  def compressed_agent_logs(self, max_bytes, top_level_dir):
    """
    Returns the last uncompressed max_bytes of the agent's logs, compressed in
    a zip file, with all log files contained in the given top_level_dir.
    """
    self._check_token()
    # In python 2.4 zipfile.py doesn't like unicode strings so we turn the
    # argument into an ascii string.
    top_level_dir = str(top_level_dir)
    file_handler = cmf.agent.get_log_rotating_file_handler()

    if not file_handler:
      msg = "Cannot find a RotatingFileHandler for the agent, is the agent " + \
          "logging to a file?"
      raise cherrypy.HTTPError(404, msg)

    compressed_log_path = None
    try:
      try:
        max_bytes = int(max_bytes)
        agent_log_compressor = AgentLogCompressor(file_handler, max_bytes, top_level_dir)
        compressed_log_path = agent_log_compressor.get_compressed_logs()
        f = self._validate_and_open(compressed_log_path, restrict_to_parent_path=False)
        return serve_fileobj(f, "application/zip")
      except ValueError:
        msg = "Invalid request. 'max_bytes' must be an integer."
        raise cherrypy.HTTPError(404, msg)
    finally:
      if compressed_log_path:
        os.remove(compressed_log_path)
  compressed_agent_logs.exposed = True

  def heartbeat(self):
    """
    Request an explicit heartbeat. The heartbeat is not returned in the body of
    the reply, but rather sent through the normal channels.
    """
    self._check_token()
    if self.wakeup_fd is None:
      raise cherrypy.HTTPError(503)
    try:
      os.write(self.wakeup_fd, "1")
    except:
      LOG.exception("Failed to write to wakeup pipe.")
      raise cherrypy.HTTPError(500)
    return "OK"
  heartbeat.exposed = True


def start_status_server(agent, host, port, wakeup_fd):
  """
  Returns the URL of the status server.
  """
  cherrypy.config.update({
    'server.socket_host': host,
    'server.socket_port': port,
    'environment': 'embedded',
    # useful for development:
    # 'request.show_tracebacks': True,
  })
  app_conf = {
    '/static': {
      'tools.staticdir.on': True,
      'tools.staticdir.root': os.path.dirname(os.path.abspath(__file__)),
      'tools.staticdir.dir': 'static',
    },
    # Note: CherryPy doesn't handle "directory indexes",
    # so it's not very useful to do this quite yet.
    '/agent_dir': {
      'tools.staticdir.on': True,
      'tools.staticdir.dir': agent.dir,
    },
  }
  # The following is much like the following quickstart,
  # but doesn't block.  Instead, we register to shut down
  # the web server at exit time.
  #   cherrypy.quickstart(StatusServer(agent), '/', config=app_conf)
  cherrypy.tree.mount(StatusServer(agent, wakeup_fd), "", config=app_conf)
  cherrypy.tree.mount(cmf.fencing_server.FencingServer(agent), "/fence", config=app_conf)
  cherrypy.engine.start()
  agent.register_for_shutdown(stop_status_server)
  return "http://%s:%s/" % (host, port)

def stop_status_server():
  cherrypy.engine.stop()
  cherrypy.engine.exit()
