// (c) Copyright 2011-2012 Cloudera, Inc. All rights reserved.
define([
	"cloudera/form/ListSelector",
	"cloudera/Util"
], function(ListSelector, Util) {

/**
 * A Mediator between the ListSelector dialog and the Chart module.
 * This controls the Add/Remove Chart feature in Service Monitoring.
 *
 * Responsibilities include:
 *
 * From the Chart module (serviceChartSet), converts the entire list of
 * time series descriptors into a set of entries that can be displayed in
 * the List Selector popup.
 *
 * Instantiates the List Selector (ListSelector) dialog.
 *
 * When user clicks the OK button, figures out which entries have been added,
 * which entries have been removed, and update the chart module.
 */
return function(options) {
	/**
	var opts = {
		'addButtonId': 'the ID of the add button',
		'entries': [],
		'updateUrl': 'the URL to update selection',
		'dialogWidth': 'the dialog width',
		'dialogHeight': 'the dialog height',
		'dialogTitle': 'the dialog title',
		'descriptionTitle': 'the description column's title',
		'serviceChartSet': the service chart set object,
		'timeSeriesDescriptors': the time series descriptors.
	};
	*/

	var serviceChartSet = options.serviceChartSet, listSelector;

	var initialChartId2Info = {};

	/**
	 * Handler called before the ListSelector dialog is opened.
	 * This populates the list of charts shown as well which ones are selected.
	 */
	var beforeOpen = function(evt) {
		jQuery.publish("pauseAutoRefresh");

		var visibleCharts = serviceChartSet.getVisibleCharts();
		populateChartId2Infos(options.timeSeriesDescriptors, visibleCharts);

		var initialChartEntries = getChartEntries();
		listSelector.setEntries(initialChartEntries);

		return true;
	};

	/**
	 * Transforms the list of time series descriptors into a
	 * list of info objects: [{name: ..., description: ..., category: ..., selected: true|false}].
	 *
	 * @param timeSeriesDescriptors - the list of time series descriptors.
	 * @param visibleCharts - the list of charts that are visible.
	 */
	var populateChartId2Infos = function(timeSeriesDescriptors, visibleCharts) {
		// Must reset.
		initialChartId2Info = {};

		// multiple time series descriptors may be mapped
		// to the same chart id.
		$.each(timeSeriesDescriptors, function(index, descriptor) {
			var id = descriptor.chart.id;
			if (initialChartId2Info.hasOwnProperty(id)) {
				initialChartId2Info[id].timeSeriesDescriptors.push(descriptor);
			} else {
				initialChartId2Info[id] = {
					name: id,
					description: descriptor.chart.title,
					category: "", // No category supported for this release.
					timeSeriesDescriptors: [descriptor],
					selected: false
				};
			}
		});

		$.each(visibleCharts, function(index, visibleChart) {
			var chartInfo = initialChartId2Info[visibleChart.id];
			if (chartInfo) {
				chartInfo.selected = true;
			}
		});
	};

	/**
	 * @return a list of chart info object, e.g.
	 * {name: ..., description: ..., category: ...,
	 * timeSeriesDescriptors: [...],
	 * selected: true|false
	 */
	var getChartEntries = function() {
		var result = [];
		$.each(initialChartId2Info, function(id, info) {
			result.push(info);
		});
		return result;
	};


	/**
	 * @return a list of added time series descriptors.
	 */
	var getAdded = function() {
		// everything that is the selected when OK is
		// pressed, but not selected on dialog open.
		var selectedIds = listSelector.getSelected();
		var predicate = function(info) {
			return !info.selected;
		};
		return mergeWithPredicate(selectedIds, predicate);
	};

	/**
	 * @return a list of removed timeSeriesDescriptors.
	 */
	var getRemoved = function() {
		// everything that is the unselected when OK is
		// pressed, but selected on dialog open.
		var unselectedIds = listSelector.getUnselected();
		var predicate = function(info) {
			return info.selected;
		};
		return mergeWithPredicate(unselectedIds, predicate);
	};

	/**
	 * @return all the time series descriptors from an array of
	 * info object, {name, description: ..., timeSeriesDescriptors: [...], }
	 * where predicate(info) is true.
	 */
	var mergeWithPredicate = function(array, predicate) {
		var result = [];

		$.each(array, function(index, id){
			if (initialChartId2Info.hasOwnProperty(id)) {
				var info = initialChartId2Info[id];
				if (predicate(info)) {
					$.merge(result, info.timeSeriesDescriptors);
				}
			}
		});

		return result;
	};

	/**
	 * Handler when user clicks the OK button.
	 */
	var afterOK = function() {
		var added = getAdded();
		var removed = getRemoved();

		$.each(removed, function(i, descriptor) {
			serviceChartSet.hideChart(descriptor.chart.id);
			serviceChartSet.hideTimeSeries(descriptor);
			// TODO:
			// currently, we cannot remove the time series.
		});

		$.each(added, function(i, descriptor) {
			serviceChartSet.addTimeSeries(descriptor);
			serviceChartSet.showChart(descriptor.chart.id);
			serviceChartSet.showTimeSeries(descriptor);
		});

		serviceChartSet.render();
	};

	var afterClose = function() {
		jQuery.publish('unpauseAutoRefresh');
	};

	/**
	 * Instantiates the ListSelector.
	 */
	var opts = {
		'id': options.addButtonId,
		'entries': [],
		'updateUrl': options.updateUrl,
		'beforeOpen': beforeOpen,
		'showOK': true,
		'showSearch': true,
		'afterOK': afterOK,
		'afterClose': afterClose,
		'dialogWidth': options.dialogWidth,
		'dialogHeight': options.dialogHeight,
		'dialogTitle': options.dialogTitle,
		'descriptionTitle': options.descriptionTitle
	};

	listSelector = new ListSelector(opts);
};
});
